---------------------
-- Marcio Buss     --
-- Anuj Maheshwari --
---------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity vga is
  port (
    clk            : in std_logic;
    pix_clk        : in std_logic;
    rst            : in std_logic;
    video_data     : in std_logic_vector(15 downto 0);
    video_addr     : out std_logic_vector(19 downto 0);
    video_req      : out std_logic;
    VIDOUT_CLK     : out std_logic;
    VIDOUT_RCR     : out std_logic_vector(9 downto 0);
    VIDOUT_GY      : out std_logic_vector(9 downto 0);
    VIDOUT_BCB     : out std_logic_vector(9 downto 0);
    VIDOUT_BLANK_N : out std_logic;
    VIDOUT_HSYNC_N : out std_logic;
    VIDOUT_VSYNC_N : out std_logic);
end vga;

architecture Behavioral of vga is

  -- Fast low-voltage TTL-level I/O pad with 12 mA drive
  
  component OBUF_F_12
    port (
      O : out STD_ULOGIC;
      I : in STD_ULOGIC);
  end component;

  -- Basic edge-sensitive flip-flop
  
  component FD
    port (
      C : in std_logic;
      D : in std_logic;
      Q : out std_logic);
  end component;

  -- Force instances of FD into pads for speed
  
  attribute iob : string;
  attribute iob of FD : component is "true";

  component vga_timing
    port (
      h_sync_delay         : out std_logic;
      v_sync_delay         : out std_logic;
      blank                : out std_logic;
      
      --RAM address generated in vga_timing.vhd
      character_ram_address : out std_logic_vector (12 downto 0);
      
      --line counter lower bits from vga_timing.vhd
      line_counter_lower_bits : out std_logic_vector (2 downto 0);
      
      -- pixel counter lower bits from vga_timing.vhd
      pixel_counter_lower_bits : out std_logic_vector(3 downto 0);
      
      pixel_clock          : in std_logic;
      reset                : in std_logic);
  end component;
  
  
  component rom
    port (
      clk  : in std_logic;
      en   : in std_logic;
      addr : in std_logic_vector(9 downto 0);
      data : out std_logic_vector(7 downto 0));
  end component;

  signal r                      : std_logic_vector (9 downto 0);
  signal g                      : std_logic_vector (9 downto 0);
  signal b                      : std_logic_vector (9 downto 0);
  signal blank                  : std_logic;
  signal hsync                  : std_logic;
  signal vsync                  : std_logic;
  
  -- RAM address has only 13 bits to address 4800 characters max
  signal character_ram_address  : std_logic_vector(12 downto 0);
  
  -- lower bits of pixel counter, coming from vga_timing module
  signal pix_cnt                : std_logic_vector(3 downto 0);
  
  -- Upper bits of ROM address. Come from RAM data
  signal rom_upper_address      : std_logic_vector(7 downto 0);
 
  -- Lower bits of ROM address. Come line counter at vga_timing module
  signal rom_lower_address      : std_logic_vector(2 downto 0);
  
  -- ROM needs to store only 128 bitmaps, which
  -- adds up to 1024 bytes: 10 bits to address
  signal rom_address            : std_logic_vector(9 downto 0);
   
  -- data from the ROM is part of a bitmap of a character
  signal rom_data               : std_logic_vector(7 downto 0);
  
  -- Counts which character from the RAM word we are accessing
  signal character_count        : std_logic_vector(1 downto 0);
  
  signal load_rom_byte          : std_logic;
  signal vga_shreg              : std_logic_vector(7 downto 0);

begin

  st : vga_timing port map (
    pixel_clock => pix_clk,
    reset => rst,
    h_sync_delay => hsync,
    v_sync_delay => vsync,
    blank => blank,
    
    -- RAM address is generated inside vga_timing module
    character_ram_address => character_ram_address,
    
    -- ROM address lower bits come from "vga_timing"
    -- module. They are the lower bits of "line_counter"
    line_counter_lower_bits => rom_lower_address,
    
    -- pixel counter generated inside vga_timing module
    pixel_counter_lower_bits => pix_cnt);
    
  rt: rom port map (
    clk => pix_clk,
    en  => load_rom_byte,
    addr => rom_address,
    data => rom_data);


  -- Load a word from the ROM at every character.
  process (pix_clk,rst)
  begin
    if rst='1' then
      load_rom_byte <= '1';
    elsif pix_clk'event and pix_clk='0' then
      if pix_cnt(2 downto 0)="111" then
         load_rom_byte <= '1';
      else
         load_rom_byte <= '0';
      end if;
    end if;
  end process;


  -- Load a word from the RAM at every TWO characters:
  -- Using the 4 least significant bits of pixel count
  -- is an alternative to do that (as opposed to "vreq")
  process (pix_clk,rst)
  begin
    if rst='1' then
      video_req <= '1';
    elsif pix_clk'event and pix_clk='0' then
      if pix_cnt="1111" then
        video_req <= '1';
      else
        video_req <= '0';
      end if;
    end if;
  end process;

  -- The video address is generated by the ram
  -- address counter in module vga_timing.vhd 

  video_addr <= "0000000" & character_ram_address(12 downto 0);
   
  -- We consider only the least significant bit of character
  -- counter. It keeps fliping between zero and one, and is
  -- used to tell us which of the two bytes from the RAM word
  -- we want to consider for the ROM address upper bits at
  -- different times.
  process (pix_clk, rst)
  begin
    if rst = '1' then
       character_count <= "00";
    elsif pix_clk'event and pix_clk = '0' then
       if pix_cnt(2 downto 0) = "111" then
         character_count <= character_count + 1;
       end if;
    end if;
  end process;
  
  
  -- Get the upper part of the ROM address from the
  -- RAM data (since RAM is word addressed, we have
  -- to load the right byte at different times
  process (pix_clk)
  begin
    if pix_clk'event and pix_clk='1' then
      if character_count(0) = '1' then
         rom_upper_address <= video_data(7 downto 0);
      else
         rom_upper_address <= video_data(15 downto 8);
      end if;
    end if;
  end process;

  rom_address(9 downto 3) <= rom_upper_address(6 downto 0);
  rom_address(2 downto 0) <= rom_lower_address;

  -- The video shift register: loads byte from the ROM
  -- at each character boundary. Shift one bit left
  -- at every pixel clock.
  process (pix_clk)
  begin
    if pix_clk'event and pix_clk='1' then
      if load_rom_byte = '1' then
        vga_shreg <= rom_data;
      else
        -- Shift one bit left. In the original circuitry,
        -- a byte corresponded to only one (color) pixel.
        -- Now, a byte corresponds to 8 pixels ON or OFF!
        for i in 0 to 6 loop
          vga_shreg(i+1) <= vga_shreg(i);
        end loop;
        vga_shreg(0) <= '0';
      end if;
    end if;
  end process;

  -- We are not using colors now. Thus, if the most significant bit
  -- on vga_shreg is '1', output RED|GREEN|BLUE. Else, output 0|0|0
  -- Now, an entire byte on vga_shreg corresponds to eight different
  -- pixels.
  r(9) <= vga_shreg (7);
  r(8) <= vga_shreg (7);
  r(7) <= vga_shreg (7);
  r(6 downto 0) <= "0000000";
  g(9) <= vga_shreg (7);
  g(8) <= vga_shreg (7);
  g(7) <= vga_shreg (7);
  g(6 downto 0) <= "0000000";
  b(9) <= vga_shreg (7);
  b(8) <= vga_shreg (7);
  b(7 downto 0) <= "00000000";

  -- Video clock I/O pad to the DAC
  
  vidclk : OBUF_F_12 port map (
    O => VIDOUT_clk,
    I => pix_clk);

  -- Control signals: hsync, vsync, and blank

  hsync_ff : FD port map (
    C => pix_clk,
    D => not hsync,
    Q => VIDOUT_HSYNC_N );

  vsync_ff : FD port map (
    C => pix_clk,
    D => not vsync,
    Q => VIDOUT_VSYNC_N );

  blank_ff :  FD port map (
    C => pix_clk,
    D => not blank,
    Q => VIDOUT_BLANK_N );

  -- Three digital color signals
  
  rgb_ff : for i in 0 to 9 generate
    
    r_ff : FD port map (
      C => pix_clk,
      D => r(i),
      Q => VIDOUT_RCR(i) );
    
    g_ff : FD port map (
      C => pix_clk,
      D => g(i),
      Q => VIDOUT_GY(i) );
    
    b_ff : FD port map (
      C => pix_clk,
      D => b(i),
      Q => VIDOUT_BCB(i) );
    
  end generate;
    
end Behavioral;
